from .simple_transformer_text_classification import SimpleTransformerTextClassificationTaskModule
from .transformer_re_text_classification import TransformerRETextClassificationTaskModule
from .transformer_seq2seq import TransformerSeq2SeqTaskModule
from .transformer_span_classification import TransformerSpanClassificationTaskModule
from .transformer_text_classification import TransformerTextClassificationTaskModule
from .transformer_token_classification import TransformerTokenClassificationTaskModule
